@@ -1,12 +1,18 @@
+from pg8000.converters import (
+    BIGINTEGER, BINARY, BOOLEAN, BOOLEAN_ARRAY, BYTES, Binary, CHAR,
+    CHAR_ARRAY, DATE, DATETIME, DECIMAL, DECIMAL_ARRAY, Date, DateFromTicks,
+    FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL,
+    JSON, JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMBER, OID, PGEnum,
+    PGInterval, PGJson, PGJsonb, PGText, PGTsvector, PGVarchar, ROWID, STRING,
+    TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, Time,
+    TimeFromTicks, Timestamp, TimestampFromTicks, UNKNOWN, UUID_TYPE, VARCHAR,
+    VARCHAR_ARRAY, XID)
 from pg8000.core import Connection, Cursor
 from pg8000.exceptions import (
-    Warning, DataError, DatabaseError, InterfaceError, ProgrammingError,
-    Error, OperationalError, IntegrityError, InternalError, NotSupportedError)
-from pg8000.converters import (
-    Binary, Date, DateFromTicks, Time, TimeFromTicks, Timestamp,
-    TimestampFromTicks, BINARY, PGInterval, PGEnum, PGJson, PGJsonb,
-    PGTsvector, PGText, PGVarchar, STRING, NUMBER, DATETIME, TIME, BOOLEAN,
-    INTEGER, BIGINTEGER, INTERVAL, JSON, JSONB, UNKNOWN, NULLTYPE, ROWID)
+    DataError, DatabaseError, Error, IntegrityError, InterfaceError,
+    InternalError, NotSupportedError, OperationalError, ProgrammingError,
+    Warning)
+
 from ._version import get_versions
 __version__ = get_versions()['version']
 del get_versions
@@ -81,5 +87,8 @@ __all__ = [
     TimeFromTicks, Timestamp, TimestampFromTicks, BINARY, PGInterval, PGEnum,
     PGJson, PGJsonb, PGTsvector, PGText, PGVarchar, STRING, NUMBER, DATETIME,
     TIME, BOOLEAN, INTEGER, BIGINTEGER, INTERVAL, JSON, JSONB, UNKNOWN,
-    NULLTYPE, ROWID
+    NULLTYPE, ROWID, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE, DECIMAL,
+    DECIMAL_ARRAY, FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER_ARRAY,
+    MACADDR, NAME, NAME_ARRAY, OID, TEXT, TEXT_ARRAY, TIMEDELTA, TIMESTAMP,
+    TIMESTAMPTZ, UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID
 ]
