@@ -1,23 +1,21 @@
-from datetime import (
-    timedelta as Timedelta, datetime as Datetime, date, time)
-from warnings import warn
 import socket
-from hashlib import md5
+from collections import defaultdict, deque
+from datetime import datetime as Datetime
 from decimal import Decimal
-from collections import deque, defaultdict
-from itertools import count, islice
 from distutils.version import LooseVersion
-import pg8000
-from scramp import ScramClient
-import enum
-from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network
-from pg8000 import converters
-from uuid import UUID
-from pg8000.exceptions import (
-    InterfaceError, ProgrammingError, Error, DatabaseError, OperationalError,
-    IntegrityError, InternalError, NotSupportedError, Warning,
-    ArrayContentNotSupportedError, ArrayDimensionsNotConsistentError)
+from hashlib import md5
+from itertools import count, islice
 from struct import Struct
+from warnings import warn
+
+import pg8000
+from pg8000 import converters
+from pg8000.exceptions import (
+    ArrayContentNotSupportedError, DatabaseError, Error, IntegrityError,
+    InterfaceError, InternalError, NotSupportedError, OperationalError,
+    ProgrammingError, Warning)
+
+from scramp import ScramClient
 
 
 def pack_funcs(fmt):
@@ -65,13 +63,6 @@ cccc_pack, cccc_unpack = pack_funcs('cccc')
 __author__ = "Mathieu Fenniak"
 
 
-NULL = i_pack(-1)
-
-'''
-FC_TEXT = 0
-FC_BINARY = 1
-'''
-
 min_int2, max_int2 = -2 ** 15, 2 ** 15
 min_int4, max_int4 = -2 ** 31, 2 ** 31
 min_int8, max_int8 = -2 ** 63, 2 ** 63
@@ -317,7 +308,7 @@ class Cursor():
     # or mapping and will be bound to variables in the operation.
     # <p>
     # Stability: Part of the DBAPI 2.0 specification.
-    def execute(self, operation, args=None, stream=None):
+    def execute(self, operation, args=(), stream=None):
         """Executes a database operation.  Parameters may be provided as a
         sequence, or as a mapping, depending upon the value of
         :data:`pg8000.paramstyle`.
@@ -345,11 +336,11 @@ class Cursor():
             .. versionadded:: 1.9.11
         """
         try:
-            self.stream = stream
-
             if not self._c.in_transaction and not self._c.autocommit:
-                self._c.execute_unnamed(self, "begin transaction", None)
-            self._c.execute_unnamed(self, operation, args, self._input_oids)
+                self._c.execute_unnamed(self, "begin transaction")
+            self._c.execute_unnamed(
+                self, operation, vals=args, input_oids=self._input_oids,
+                stream=stream)
             self._input_oids = None
         except AttributeError as e:
             if self._c is None:
@@ -465,7 +456,10 @@ class Cursor():
             if isinstance(size, int):
                 oid = size
             else:
-                oid, _ = self._c.py_types[size]
+                try:
+                    oid, _ = self._c.py_types[size]
+                except KeyError:
+                    oid = pg8000.converters.UNKNOWN
             oids.append(oid)
 
         self._input_oids = oids
@@ -514,6 +508,7 @@ EMPTY_QUERY_RESPONSE = b"I"
 
 BIND = b"B"
 PARSE = b"P"
+QUERY = b"Q"
 EXECUTE = b"E"
 FLUSH = b'H'
 SYNC = b'S'
@@ -557,9 +552,6 @@ IDLE_IN_TRANSACTION = b"T"
 IDLE_IN_FAILED_TRANSACTION = b"E"
 
 
-arr_trans = dict(zip(map(ord, "[] '"), list('{}') + [None] * 2))
-
-
 class Connection():
 
     # DBAPI Extension: supply exceptions as attributes on the connection
@@ -694,79 +686,8 @@ class Connection():
         self._backend_key_data = None
 
         self.pg_types = defaultdict(
-            lambda: converters.text_in, {
-                16: converters.bool_in,  # boolean
-                17: converters.bytea_in,  # bytea
-                19: converters.text_in,  # name type
-                20: int,  # int8
-                21: int,  # int2
-                22: converters.vector_in,  # int2vector
-                23: int,  # int4
-                25: converters.text_in,  # TEXT type
-                26: int,  # oid
-                28: int,  # xid
-                114: converters.json_in,  # json
-                700: float,  # float4
-                701: float,  # float8
-                705: converters.text_in,  # unknown
-                829: converters.text_in,  # MACADDR type
-                869: converters.inet_in,  # inet
-                1000: converters.array_bool_in,  # BOOL[]
-                1003: converters.array_text_in,  # NAME[]
-                1005: converters.array_int_in,  # INT2[]
-                1007: converters.array_int_in,  # INT4[]
-                1009: converters.array_text_in,  # TEXT[]
-                1014: converters.array_text_in,  # CHAR[]
-                1015: converters.array_text_in,  # VARCHAR[]
-                1016: converters.array_int_in,  # INT8[]
-                1021: converters.array_float_in,  # FLOAT4[]
-                1022: converters.array_float_in,  # FLOAT8[]
-                1042: converters.text_in,  # CHAR type
-                1043: converters.text_in,  # VARCHAR type
-                1082: converters.date_in,  # date
-                1083: converters.time_in,
-                1114: converters.timestamp_in,
-                1184: converters.timestamptz_in,  # timestamp w/ tz
-                1186: converters.timedelta_in,
-                1231: converters.array_numeric_in,  # NUMERIC[]
-                1263: converters.array_text_in,  # cstring[]
-                1700: converters.numeric_in,  # NUMERIC
-                2275: converters.text_in,  # cstring
-                2950: converters.uuid_in,  # uuid
-                3802: converters.json_in,  # jsonb
-            }
-        )
-
-        self.py_types = {
-            type(None): (-1, converters.null_out),  # null
-            bool: (16, converters.bool_out),
-            bytearray: (17, converters.bytes_out),  # bytea
-            20: (20, converters.int_out),  # int8
-            21: (21, converters.int_out),  # int2
-            23: (23, converters.int_out),  # int4
-            converters.PGText: (25, converters.text_out),  # text
-            float: (701, converters.float_out),  # float8
-            converters.PGEnum: (705, converters.enum_out),
-            date: (1082, converters.date_out),  # date
-            time: (1083, converters.time_out),  # time
-            1114: (1114, converters.timestamp_out),  # timestamp
-            converters.PGVarchar: (1043, converters.text_out),  # varchar
-            1184: (1184, converters.timestamptz_out),  # timestamptz
-            converters.PGJson: (114, converters.text_out),
-            converters.PGJsonb: (3802, converters.text_out),
-            Timedelta: (1186, converters.timedelta_out),
-            converters.PGInterval: (1186, converters.pginterval_out),
-            Decimal: (1700, converters.numeric_out),  # Decimal
-            converters.PGTsvector: (3614, converters.text_out),
-            UUID: (2950, converters.uuid_out),  # uuid
-            bytes: (17, converters.bytes_out),  # bytea
-            str: (705, converters.text_out),  # unknown
-            enum.Enum: (705, converters.enum_out),
-            IPv4Address: (869, converters.inet_out),  # inet
-            IPv6Address: (869, converters.inet_out),  # inet
-            IPv4Network: (869, converters.inet_out),  # inet
-            IPv6Network: (869, converters.inet_out),  # inet
-        }
+            lambda: converters.text_in, converters.PG_TYPES)
+        self.py_types = dict(converters.PY_TYPES)
 
         self.inspect_funcs = {
             Datetime: self.inspect_datetime,
@@ -939,13 +860,11 @@ class Connection():
         # guaranteed for v1.xx.
         backend_pid = i_unpack(data)[0]
         idx = 4
-        null = data.find(NULL_BYTE, idx) - idx
-        condition = data[idx:idx + null].decode("ascii")
-        idx += null + 1
-        null = data.find(NULL_BYTE, idx) - idx
-        # additional_info = data[idx:idx + null]
+        null_idx = data.find(NULL_BYTE, idx)
+        channel = data[idx:null_idx].decode("ascii")
+        payload = data[null_idx+1:-1].decode('ascii')
 
-        self.notifications.append((backend_pid, condition))
+        self.notifications.append((backend_pid, channel, payload))
 
     def cursor(self):
         """Creates a :class:`Cursor` object bound to this
@@ -970,7 +889,7 @@ class Connection():
         This function is part of the `DBAPI 2.0 specification
         <http://www.python.org/dev/peps/pep-0249/>`_.
         """
-        self.execute_unnamed(self._cursor, "commit", None)
+        self.execute_unnamed(self._cursor, "commit")
 
     def rollback(self):
         """Rolls back the current database transaction.
@@ -980,7 +899,7 @@ class Connection():
         """
         if not self.in_transaction:
             return
-        self.execute_unnamed(self._cursor, "rollback", None)
+        self.execute_unnamed(self._cursor, "rollback")
 
     def close(self):
         """Closes the database connection.
@@ -1106,14 +1025,14 @@ class Connection():
             oid = 25
             # Use binary ARRAY format to avoid having to properly
             # escape text in the array literals
-            array_oid = pg_array_types[oid]
+            array_oid = converters.PG_ARRAY_TYPES[oid]
         else:
             # supported array output
             typ = type(first_element)
 
             if typ == bool:
                 oid = 16
-                array_oid = pg_array_types[oid]
+                array_oid = converters.PG_ARRAY_TYPES[oid]
 
             elif issubclass(typ, int):
                 # special int array support -- send as smallest possible array
@@ -1150,7 +1069,7 @@ class Connection():
                         oid = 25
                         # Use binary ARRAY format to avoid having to properly
                         # escape text in the array literals
-                    array_oid = pg_array_types[oid]
+                    array_oid = converters.PG_ARRAY_TYPES[oid]
                 except KeyError:
                     raise ArrayContentNotSupportedError(
                         "oid " + str(oid) + " not supported as array contents")
@@ -1241,6 +1160,7 @@ class Connection():
         count = h_unpack(data)[0]
         idx = 2
         row_desc = []
+        input_funcs = []
         for i in range(count):
             name = data[idx:data.find(NULL_BYTE, idx)]
             idx += len(name) + 1
@@ -1251,8 +1171,9 @@ class Connection():
             field['name'] = name
             idx += 18
             row_desc.append(field)
-            field['func'] = self.pg_types[field['type_oid']]
+            input_funcs.append(self.pg_types[field['type_oid']])
         cursor.row_desc = row_desc
+        cursor.input_funcs = input_funcs
 
     def send_PARSE(self, statement_name_bin, statement, oids):
 
@@ -1267,41 +1188,60 @@ class Connection():
     def send_DESCRIBE_STATEMENT(self, statement_name_bin):
         self._send_message(DESCRIBE, STATEMENT + statement_name_bin)
 
-    def execute_unnamed(self, cursor, operation, vals, input_oids=None):
-        if vals is None:
-            vals = ()
-
-        statement, make_args = convert_paramstyle(cursor.paramstyle, operation)
-
-        args = make_args(vals)
-        param_oids, params = self.make_params(args)
-        oids = param_oids if input_oids is None else input_oids
-
-        statement_name_bin = NULL_BYTE
-
-        self.send_PARSE(statement_name_bin, statement, oids)
-        self.send_DESCRIBE_STATEMENT(statement_name_bin)
-        self._write(SYNC_MSG)
-
+    def send_QUERY(self, sql):
+        data = sql.encode(self._client_encoding) + NULL_BYTE
         try:
-            self._flush()
-        except AttributeError as e:
-            if self._sock is None:
+            self._write(QUERY)
+            self._write(i_pack(len(data) + 4))
+            self._write(data)
+        except ValueError as e:
+            if str(e) == "write to closed file":
                 raise InterfaceError("connection is closed")
             else:
                 raise e
+        except AttributeError:
+            raise InterfaceError("connection is closed")
 
-        self.handle_messages(cursor)
-
-        cursor.input_funcs = tuple(f['func'] for f in cursor.row_desc)
-
+    def execute_unnamed(
+            self, cursor, operation, vals=(), input_oids=None, stream=None):
         cursor._cached_rows.clear()
         cursor._row_count = -1
-        self.send_BIND(statement_name_bin, params)
-        self.send_EXECUTE()
-        self._write(SYNC_MSG)
-        self._flush()
-        self.handle_messages(cursor)
+        cursor.stream = stream
+        if len(vals) == 0 and cursor.stream is None:
+            cursor.row_desc = []
+            self.send_QUERY(operation)
+            self._flush()
+            self.handle_messages(cursor)
+        else:
+            statement, make_args = convert_paramstyle(
+                cursor.paramstyle, operation)
+
+            param_oids, params = self.make_params(make_args(vals))
+            oids = param_oids if input_oids is None else input_oids
+
+            self.send_PARSE(NULL_BYTE, statement, oids)
+            self._write(SYNC_MSG)
+            self._flush()
+            self.handle_messages(cursor)
+            self.send_DESCRIBE_STATEMENT(NULL_BYTE)
+
+            self._write(SYNC_MSG)
+
+            try:
+                self._flush()
+            except AttributeError as e:
+                if self._sock is None:
+                    raise InterfaceError("connection is closed")
+                else:
+                    raise e
+
+            self.send_BIND(NULL_BYTE, params)
+            self.handle_messages(cursor)
+            self.send_EXECUTE()
+
+            self._write(SYNC_MSG)
+            self._flush()
+            self.handle_messages(cursor)
 
     def prepare(self, operation):
         return PreparedStatement(self, operation)
@@ -1328,12 +1268,10 @@ class Connection():
             else:
                 raise e
 
-        self.handle_messages(self._run_cursor)
+        cursor = self._run_cursor
+        self.handle_messages(cursor)
 
-        row_desc = self._run_cursor.row_desc
-        input_funcs = tuple(f['func'] for f in row_desc)
-
-        return statement_name_bin, row_desc, input_funcs
+        return statement_name_bin, cursor.row_desc, cursor.input_funcs
 
     def execute_named(self, statement_name_bin, params, row_desc, input_funcs):
         cursor = self._run_cursor
@@ -1468,7 +1406,8 @@ class Connection():
         self.parameter_statuses.append((key, value))
         if key == b"client_encoding":
             encoding = value.decode("ascii").lower()
-            self._client_encoding = pg_to_py_encodings.get(encoding, encoding)
+            self._client_encoding = converters.pg_to_py_encodings.get(
+                encoding, encoding)
 
         elif key == b"integer_datetimes":
             if value == b'on':
@@ -1481,7 +1420,7 @@ class Connection():
             self._server_version = LooseVersion(value.decode('ascii'))
             if self._server_version < LooseVersion('8.2.0'):
                 self._commands_with_count = (
-                    b"INSERT", b"DELETE", b"UPDATE", b"MOVE", b"FETCH")
+                    b"INSERT", b"DELETE", b"UPDATE", b"MOVE")
             elif self._server_version < LooseVersion('9.0.0'):
                 self._commands_with_count = (
                     b"INSERT", b"DELETE", b"UPDATE", b"MOVE", b"FETCH",
@@ -1510,7 +1449,7 @@ class Connection():
         """
         self._xid = xid
         if self.autocommit:
-            self.execute(self._cursor, "begin transaction", None)
+            self.execute_unnamed(self._cursor, "begin transaction")
 
     def tpc_prepare(self):
         """Performs the first phase of a transaction started with .tpc_begin().
@@ -1524,7 +1463,7 @@ class Connection():
         <http://www.python.org/dev/peps/pep-0249/>`_.
         """
         q = "PREPARE TRANSACTION '%s';" % (self._xid[1],)
-        self.execute(self._cursor, q, None)
+        self.execute_unnamed(self._cursor, q)
 
     def tpc_commit(self, xid=None):
         """When called with no arguments, .tpc_commit() commits a TPC
@@ -1555,9 +1494,8 @@ class Connection():
             previous_autocommit_mode = self.autocommit
             self.autocommit = True
             if xid in self.tpc_recover():
-                self.execute(
-                    self._cursor, "COMMIT PREPARED '%s';" % (xid[1], ),
-                    None)
+                self.execute_unnamed(
+                    self._cursor, "COMMIT PREPARED '%s';" % (xid[1], ))
             else:
                 # a single-phase commit
                 self.commit()
@@ -1591,9 +1529,8 @@ class Connection():
             self.autocommit = True
             if xid in self.tpc_recover():
                 # a two-phase rollback
-                self.execute(
-                    self._cursor, "ROLLBACK PREPARED '%s';" % (xid[1],),
-                    None)
+                self.execute_unnamed(
+                    self._cursor, "ROLLBACK PREPARED '%s';" % (xid[1],))
             else:
                 # a single-phase rollback
                 self.rollback()
@@ -1618,16 +1555,6 @@ class Connection():
             self.autocommit = previous_autocommit_mode
 
 
-# pg element oid -> pg array typeoid
-pg_array_types = {
-    16: 1000,
-    25: 1009,    # TEXT[]
-    701: 1022,
-    1043: 1009,
-    1700: 1231,  # NUMERIC[]
-}
-
-
 def array_find_first_element(arr):
     for v in array_flatten(arr):
         if v is not None:
@@ -1644,92 +1571,6 @@ def array_flatten(arr):
             yield v
 
 
-def array_check_dimensions(arr):
-    if len(arr) > 0:
-        v0 = arr[0]
-        if isinstance(v0, list):
-            req_len = len(v0)
-            req_inner_lengths = array_check_dimensions(v0)
-            for v in arr:
-                inner_lengths = array_check_dimensions(v)
-                if len(v) != req_len or inner_lengths != req_inner_lengths:
-                    raise ArrayDimensionsNotConsistentError(
-                        "array dimensions not consistent")
-            retval = [req_len]
-            retval.extend(req_inner_lengths)
-            return retval
-        else:
-            # make sure nothing else at this level is a list
-            for v in arr:
-                if isinstance(v, list):
-                    raise ArrayDimensionsNotConsistentError(
-                        "array dimensions not consistent")
-    return []
-
-
-def array_has_null(arr):
-    for v in array_flatten(arr):
-        if v is None:
-            return True
-    return False
-
-
-# PostgreSQL encodings:
-#   http://www.postgresql.org/docs/8.3/interactive/multibyte.html
-# Python encodings:
-#   http://www.python.org/doc/2.4/lib/standard-encodings.html
-#
-# Commented out encodings don't require a name change between PostgreSQL and
-# Python.  If the py side is None, then the encoding isn't supported.
-pg_to_py_encodings = {
-    # Not supported:
-    "mule_internal": None,
-    "euc_tw": None,
-
-    # Name fine as-is:
-    # "euc_jp",
-    # "euc_jis_2004",
-    # "euc_kr",
-    # "gb18030",
-    # "gbk",
-    # "johab",
-    # "sjis",
-    # "shift_jis_2004",
-    # "uhc",
-    # "utf8",
-
-    # Different name:
-    "euc_cn": "gb2312",
-    "iso_8859_5": "is8859_5",
-    "iso_8859_6": "is8859_6",
-    "iso_8859_7": "is8859_7",
-    "iso_8859_8": "is8859_8",
-    "koi8": "koi8_r",
-    "latin1": "iso8859-1",
-    "latin2": "iso8859_2",
-    "latin3": "iso8859_3",
-    "latin4": "iso8859_4",
-    "latin5": "iso8859_9",
-    "latin6": "iso8859_10",
-    "latin7": "iso8859_13",
-    "latin8": "iso8859_14",
-    "latin9": "iso8859_15",
-    "sql_ascii": "ascii",
-    "win866": "cp886",
-    "win874": "cp874",
-    "win1250": "cp1250",
-    "win1251": "cp1251",
-    "win1252": "cp1252",
-    "win1253": "cp1253",
-    "win1254": "cp1254",
-    "win1255": "cp1255",
-    "win1256": "cp1256",
-    "win1257": "cp1257",
-    "win1258": "cp1258",
-    "unicode": "utf-8",  # Needed for Amazon Redshift
-}
-
-
 class PreparedStatement():
     def __init__(self, con, operation):
         self.con = con
@@ -1739,9 +1580,7 @@ class PreparedStatement():
 
     def run(self, **vals):
 
-        args = self.make_args(vals)
-        oids, params = self.con.make_params(args)
-
+        oids, params = self.con.make_params(self.make_args(vals))
         cursor = self.con._run_cursor
 
         try:
@@ -1752,7 +1591,7 @@ class PreparedStatement():
 
         try:
             if not self.con.in_transaction and not self.con.autocommit:
-                self.con.execute_unnamed(cursor, "begin transaction", None)
+                self.con.execute_unnamed(cursor, "begin transaction")
             self.con.execute_named(name_bin, params, row_desc, input_funcs)
         except AttributeError as e:
             if self.con is None:
